package edu.sdsc.inca.consumer;

import junit.framework.TestCase;
import org.apache.log4j.Logger;
import edu.sdsc.inca.ConsumerTest;
import edu.sdsc.inca.Consumer;

import java.io.File;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

/**
 * Tests the CachedQuery class.
 *
 * @author Shava Smallen &lt;ssmallen@sdsc.edu&gt;
 */
public class CachedQueryTest extends TestCase {
  private static Logger logger = Logger.getLogger( CachedQueryTest.class );

  /**
   * Verify the result is not null and contains a timestamp and return it.
   *
   * @param result A string containing the result of a query.
   *
   * @return  The timestamp retrieved from the result.
   */
  public static long getAndVerifyResult( String result ) {
    logger.info( "Received " + result );
    assertNotNull( "Found first result", result );
    Pattern pattern = Pattern.compile( "<time>(\\d+)</time>" );
    Matcher matcher = pattern.matcher( result );
    assertTrue( "got time string from first query", matcher.find() );
    return Long.parseLong( matcher.group(1) );
  }
  
  /**
   * Test caching abilities.
   *
   * @throws Exception
   */
  public void testCache() throws Exception {
    File qFile = new File( "var/queries/q1.xml" );
    if ( qFile.exists() ) qFile.delete();

    ConsumerTest.ConsumerTester tester = new ConsumerTest.ConsumerTester(1, 4);
    tester.consumer.setCacheReloadPeriod( 20000 );
    tester.start();

    // Before we start the query thread, there should be no result in the cache
    CachedQuery q1 = new CachedQuery( "q1", "select guid from suite", 5 );
    String r1 = q1.getResult();
    assertNull( "No result in cache yet", r1 );

    // We start the cache thread and wait a few seconds and verify that the
    // thread queried the depot and stored the result
    logger.info ( "Test 1 started" );
    q1.start();
    logger.info( "query thread started" );
    long firstQueryTime = getAndVerifyResult( q1.getResultOrWait() );


    // We do another query and verify the query result got updated by
    // comparing timestamps
    logger.info ( "Test 2 started" );
    Thread.sleep( 6000 );
    long secondQueryTime = getAndVerifyResult( q1.getResult() );
    assertTrue( "second query time later than first",
                secondQueryTime > firstQueryTime );


    // We stop the current cache thread and start a new one to emulate a
    // a consumer restart -- when it first starts, it should return the
    // the results from file because they're still fresh
    logger.info ( "Test 3 started" );
    q1.interrupt();
    q1.join();
    CachedQuery q2 = new CachedQuery( "q1", "select guid from suite", 5 );
    q2.start();
    Thread.sleep( 1000 );
    long thirdQueryTime = getAndVerifyResult( q2.getResult() );
    assertTrue( "3rd query time is same as 2nd",
                thirdQueryTime == secondQueryTime );

    // now let's make sure, it updates properly
    Thread.sleep( 5000 );
    long fourthQueryTime = getAndVerifyResult( q2.getResult() );
    assertTrue( "4th query time is greater than 3rd",
                fourthQueryTime > thirdQueryTime );

    q2.interrupt();
    logger.info( "Shutting down now" );
    tester.stop();
    logger.info( "Shut down complete" );
  }

  /**
   * Test filename translation
   *
   * @throws Exception
   */
  public void testFile() throws Exception {

    Consumer.getGlobalConsumer().setTempPath( "var" );
    CachedQuery q1 = new CachedQuery( "q1", "select guid from suite", 5 );
    assertEquals( "regular filename", "q1.xml", q1.getCacheFile().getName() );

    q1.setQueryName( "My query is cool" );
    assertEquals( "spaces in filename",
                  "My_query_is_cool.xml", q1.getCacheFile().getName() );

    q1.setQueryName( "My  query is cool" );
    assertEquals( "two spaces in filename",
                  "My__query_is_cool.xml", q1.getCacheFile().getName() );
  }

}
